use crate::formatter::format::ByteOrderOperations;
use bitvec::prelude::*;
use std::fmt::{Display, Formatter};

#[derive(Clone, Debug)]
pub enum ByteStreamError {
    OutOfRange,
}

impl Display for ByteStreamError {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            ByteStreamError::OutOfRange => write!(f, "Requested values out of range"),
        }
    }
}

pub const fn bit_mask(mask: u8) -> u8 {
    match mask {
        0 => 0x00,
        1 => 0x01,
        2 => 0x03,
        3 => 0x07,
        4 => 0x0f,
        5 => 0x1f,
        6 => 0x3f,
        7 => 0x7f,
        _ => 0xff,
    }
}

#[derive(Default, Clone, Debug)]
pub struct ByteStream {
    data: Vec<u8>,
    reverse_bits: bool,
}

impl ByteStream {
    pub fn set_reverse_bits(&mut self, reverse_bits: bool) {
        self.reverse_bits = reverse_bits;
    }

    pub fn get_reverse_bits(&self) -> bool {
        self.reverse_bits
    }

    pub fn get_bytes<T: ByteOrderOperations>(
        &self,
        mut bit_ndx: usize,
        bit_count: usize,
    ) -> Result<Vec<u8>, ByteStreamError> {
        let mut bits = self.data.view_bits::<Msb0>().to_bitvec();

        if bit_count + bit_ndx > bits.len() {
            return Err(ByteStreamError::OutOfRange);
        }

        if self.reverse_bits {
            bits.reverse();
        }

        let mut data: Vec<u8> = vec![];

        if bit_count % 8 != 0 {
            let bits_needed = bit_count % 8;
            data.push(bits[bit_ndx..bit_ndx + bits_needed].load::<u8>());
            bit_ndx += bits_needed;
        }

        for _ in 0..bit_count / 8 {
            data.push(bits[bit_ndx..bit_ndx + 8].load::<u8>());
            bit_ndx += 8;
        }

        Ok(data)
    }

    pub fn len(&self) -> usize {
        self.data.len()
    }
}

impl From<&[u8]> for ByteStream {
    fn from(slice: &[u8]) -> Self {
        ByteStream::from(slice.to_vec())
    }
}

impl From<Vec<u8>> for ByteStream {
    fn from(vec: Vec<u8>) -> Self {
        ByteStream {
            data: vec,
            reverse_bits: false,
        }
    }
}

#[cfg(test)]
mod tests {
    use super::ByteStream;
    use byteorder::{BigEndian, LittleEndian};

    #[test]
    fn test_get_bytes_no_shift() {
        let bytes: Vec<u8> = vec![0xff, 0x00, 0x55];
        let bit_stream = ByteStream::from(bytes.clone());

        let new_bytes = bit_stream
            .get_bytes::<BigEndian>(0, bytes.len() * 8)
            .unwrap();
        assert_eq!(bytes, new_bytes);
    }

    #[test]
    fn test_get_bytes_with_shift_in_byte() {
        let bytes: Vec<u8> = vec![0x5f, 0x00, 0x55];
        let bit_stream = ByteStream::from(bytes.clone());

        let new_bytes = bit_stream.get_bytes::<BigEndian>(4, 4).unwrap();
        assert_eq!(vec![0x0f], new_bytes);
    }

    #[test]
    fn test_get_bytes_with_shift_across_bytes() {
        let bytes: Vec<u8> = vec![0xff, 0x55];
        let bit_stream = ByteStream::from(bytes.clone());

        let new_bytes = bit_stream.get_bytes::<LittleEndian>(4, 8).unwrap();
        assert_eq!(vec![0x5f], new_bytes);
    }

    #[test]
    fn test_get_bytes_with_shift_across_bytes_odd() {
        let bytes: Vec<u8> = vec![0xff, 0x55];
        let bit_stream = ByteStream::from(bytes.clone());

        let new_bytes = bit_stream.get_bytes::<LittleEndian>(7, 2).unwrap();
        assert_eq!(vec![0x01], new_bytes);
    }

    #[test]
    fn test_get_bytes_one_byte() {
        let bytes: Vec<u8> = vec![0xff, 0x55];
        let bit_stream = ByteStream::from(bytes.clone());

        let new_bytes = bit_stream.get_bytes::<BigEndian>(8, 8).unwrap();
        assert_eq!(vec![0x55], new_bytes);
    }

    #[test]
    fn test_get_bytes_3_bits() {
        let bytes: Vec<u8> = vec![0x55];
        let bit_stream = ByteStream::from(bytes.clone());

        let new_bytes = bit_stream.get_bytes::<BigEndian>(0, 3).unwrap();
        assert_eq!(vec![0x02], new_bytes);
    }
}
